/*****************************************************************************
 * Multithreaded Web Client
 *
 * Copyright  2002 by Beth Plale <plale@cs.indiana.edu>. All rights reserved.
 * URL: http://www.cs.indiana.edu/~plale
 *
 * A simple multithreaded client for issuing GET requests to a web server.
 * Uses the HTTP 1.0 protocol, TCP sockets, and POSIX Pthreads.
 *
 * This code is not to be distributed freely; it is for personal use
 * only during P436 coursework in Computer Science at Indiana University.
 * All other uses must be approved by the author.
 *
 *****************************************************************************
 * Modified: 11/14/02 Craig M. Jacobs and Charles W. Moad
 * Added proxy support, will not work with standard HTTP requests
 * must use with a proxy
 *****************************************************************************
 * Modified: 11/16/02 John David Ratliff
 * Added Linux support (replaced getipnodebyname with getaddrinfo)
 * Changed program to continue on error instead of exit
 * Changed bytes from int to long long (may require gcc to compile, now)
 * Changed http request not to use 0, since that segfaults on Solaris
 *         since the array was never initialized at 0.
 * Changed to work with or without a proxy server.
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <ctype.h>

#include "client.h"
#include "FTimer.h"

#define NUM_CLIENTS 10
#define NUM_FILES_TO_GET 100
#define MAX_REQ_LEN 512
#define MAX_ADDR_LEN 256

#define HTTPD_ADDR "127.0.0.1"
#define HTTPD_PORT 8080
#define PROXY_ADDR "127.0.0.1"
#define PROXY_PORT 8081

extern int h_errno;

pthread_mutex_t global_vars;
unsigned int num_files_to_get;     /* READ ONLY, set once in main() */
pthread_t* clients; 		   /* Arrary of client pthread id's */
int proxy_port;                    /* global variable */
char * proxy_addr;
int http_port;                     /* global variable */
char * http_addr;

// set to true if we should use a proxy...
short int useProxy = 0;

unsigned long long int global_bytes = 0;

// returns the network byte ordered IP address of the host 'hostname'
// this one is thread-safe now!
in_addr_t get_ip_address(const char *hostname) {
    struct addrinfo *res = NULL;
    struct sockaddr_in *address;
    int error;
    in_addr_t ip = 0;

    error = getaddrinfo(hostname,NULL,NULL,&res);

    if (error == 0) {
        address = (struct sockaddr_in *)res->ai_addr;
        ip = address->sin_addr.s_addr;
        freeaddrinfo(res);
    }

    return ip;
}

/*
 * Thread runs to connect to client, loops forever.
 */
void *client_tester( void* unused )
{
    int sock, recv_bytes=0, amt, req_len, file;
    unsigned int cnt;
    double fileIndex;
    char buf[MAX_FILE_LEN], req_str[MAX_REQ_LEN];
    char fullserveraddr[MAX_ADDR_LEN];
    struct sockaddr_in sin;
    pthread_t tid;
    tid = pthread_self();

    for (cnt = 0; cnt < num_files_to_get; cnt++) {
	
	/* Create socket for incoming connections */
	//	if ((sock = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0){
	if ((sock = socket(AF_INET,SOCK_STREAM,0)) < 0) {
	    perror("Error Getting Socket");
	    continue;
	}

	/* construct local address structure */
	memset(&sin,0,sizeof(sin));
	sin.sin_family = AF_INET;

	/*	
	// thread safe and unice independent...
	sin.sin_addr.s_addr = get_ip_address(proxy_addr);
	sin.sin_port = htons(proxy_port);
	*/

	if (useProxy) {
	    sin.sin_addr.s_addr = get_ip_address(proxy_addr);
	    sin.sin_port = htons(proxy_port);
	} else {
	    sin.sin_addr.s_addr = get_ip_address(http_addr);
	    sin.sin_port = htons(http_port);
	}

	if ( connect(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	    fprintf(stderr,"Bad connect to %s at %d\n", proxy_addr, proxy_port);
	    perror("unable to connect - continuing...\n");
	    close(sock);
	    continue;
	}
	
	snprintf(fullserveraddr, MAX_ADDR_LEN - 1, "http://%s:%d", http_addr, http_port);

	/* Generate a random real and get file according to distribution function
	 * did not accept drand... maybe no library support for drand
	 */
	fileIndex = (double) ((1000*rand()/(RAND_MAX+1.0))/1000);
	printf("\nfileindex is: %f\n",fileIndex);

	if(fileIndex >= 0 && fileIndex <= 0.015) {   /* small files */
	    //	    req_len = snprintf(req_str,MAX_REQ_LEN-1,"GET %s/%s HTTP/1.0\r\n\r\n",
	    //	       fullserveraddr,
	    //	       http_files[(int) ((15.0*rand()/(RAND_MAX+1.0))+1)]);
	    file = 1 + (int) ((15.0*rand()/(RAND_MAX+1.0)));
	} else if(fileIndex > 0.015 && fileIndex <= 0.765) {  /* 1.5K - 10K files */
	    //	    req_len = snprintf(req_str,MAX_REQ_LEN-1,"GET %s/%s HTTP/1.0\r\n\r\n",
	    //	       fullserveraddr,
	    //	       http_files[15 + (int) (750.0*rand()/(RAND_MAX+1.0))]);
	    file = 16 + (int) (750.0*rand()/(RAND_MAX+1.0));
	} else if(fileIndex > 0.765 && fileIndex <= 0.915) {
	    //	    req_len = snprintf(req_str,MAX_REQ_LEN-1,"GET %s/%s HTTP/1.0\r\n\r\n",
	    //	       fullserveraddr,
	    //	       http_files[765 + (int) (150.0*rand()/(RAND_MAX+1.0))]);
	    file = 766 + (int) (150.0*rand()/(RAND_MAX+1.0));
	} else if(fileIndex > 0.915 && fileIndex <= 0.990) {
	    //	    req_len = snprintf(req_str,MAX_REQ_LEN-1,"GET %s/%s HTTP/1.0\r\n\r\n",
	    //	       fullserveraddr,
	    //	       http_files[915 + (int) (75.0*rand()/(RAND_MAX+1.0))]);
	    file = 916 + (int) (75.0*rand()/(RAND_MAX+1.0));
	} else {   /* large files */
	    //	    req_len = snprintf(req_str,MAX_REQ_LEN-1,"GET %s/%s HTTP/1.0\r\n\r\n",
	    //	       fullserveraddr,
	    //	       http_files[990 + (int) (10.0*rand()/(RAND_MAX+1.0))]);
	    file = 991 + (int)(10.0*rand()/(RAND_MAX+1.0));
	}

	if (useProxy) {
	    req_len = snprintf(req_str, MAX_REQ_LEN - 1, "GET %s/%s HTTP/1.0\r\n\r\n",
			       fullserveraddr,
			       http_files[file]);
	} else {
	    req_len = snprintf(req_str, MAX_REQ_LEN - 1, "GET /%s HTTP/1.0\r\n\r\n",
			       http_files[file]);
	}

	/* Send request */
	if ( send(sock,req_str,req_len,0) != req_len ) {
	    printf("Send did not send all of data, continuing...\n");
	    close(sock);
	    continue;
	}
	
	/*
	 * Because we keep track of recv_bytes only, no need to reinitialize
	 * buffer.
	 */
	
	while ((amt = recv(sock, buf, MAX_FILE_LEN - 1, 0)) > 0) {
	    recv_bytes += amt;
	}
	
	if (amt < 0) {
	    perror("Bad recv(), continuing...\n");
	    close(sock);
	    continue;
	}

	fprintf(stderr,"[TID %d, req  %d]: received %d bytes\n",(int)tid, cnt, recv_bytes);
	close( sock );

	pthread_mutex_lock(&global_vars);
	global_bytes += recv_bytes;
	pthread_mutex_unlock(&global_vars);

	recv_bytes = 0;
    }
    
    return NULL;
}

void http_files_init()
{
    int i,j,k[4],m,p,index;
    m=0;

    for(i=1;i<=1000;i++)
	http_files[i] = (char *) malloc(sizeof(char *) * 30);

    for(index=1;index<=1000;index++) {
	i = index;
	while(i!=0) {
	    j = i%10;
	    k[m++]=j;
	    i = i/10;
	}
	for(p=m-1;p>=0;p--) {
	    sprintf(http_files[index],"%s%d",http_files[index],toascii(k[p]));
	}
	m=0;
	strcat(http_files[index],".html");
    }

}

int main( int argc, char* argv[] )
{
    FTR_wall_timer_val start_time, end_time, interval_time;
    double total_seconds = 0.0;
    unsigned int num_clients_to_use;
    unsigned int i;
    double MBps;

    if (((argc != 5) && (argc != 7)) || (argv[1][0] == '-' && argv[1][1] == 'h')) {
	printf("Usage: client <num threads> <num files to get> [proxy_name] [proxy_port] <http_name> <http_port>\n");
	exit(0);
    }

    num_clients_to_use = (argc >= 2) ? atoi(argv[1]) : NUM_CLIENTS;
    num_files_to_get = (argc >= 3) ? atoi(argv[2]) : NUM_FILES_TO_GET;

    if (argc == 7) {
	useProxy = 1;
	proxy_addr = argv[3];
	proxy_port = atoi(argv[4]);
	http_addr = argv[5];
	http_port = atoi(argv[6]);
    } else if (argc == 5) {
	useProxy = 0;
	http_addr = argv[3];
	http_port = atoi(argv[4]);
    }

    //    proxy_addr = (argc >= 4) ? argv[3] : PROXY_ADDR;
    //    proxy_port = (argc == 7) ? atoi(argv[4]) : PROXY_PORT;
    //    http_addr = (argc >= 6) ? argv[5] : HTTPD_ADDR;
    //    http_port = (argc == 7) ? atoi(argv[6]) : HTTPD_PORT;

    /* initialize all files */
    http_files_init();
    srand(getpid());
    if ( (clients = (pthread_t *) malloc(num_clients_to_use*sizeof(*clients)))
	 == NULL){
	perror("malloc failed, exiting!\n");
	exit(1);
    }
    FTR_Get_wall_time(start_time);
    for (i=0; i<num_clients_to_use; i++) {
	pthread_create(&(clients[i]),NULL,client_tester,NULL);
    }
    for (i=0; i<num_clients_to_use; i++) {
	pthread_join(clients[i],NULL);
    }
    FTR_Get_wall_time(end_time);
    FTR_Get_wall_interval(start_time, end_time, interval_time);
    total_seconds = FTR_Seconds(interval_time);
    MBps = (global_bytes / total_seconds)/1000000;
    printf("Total bytes of %llu in %.3f seconds ",global_bytes,total_seconds);
    printf("\nis throughput of %f MBps\n", MBps);
    return(0);
}

