/*************************************************************
 * CSCI P436 - Operating Systems
 *
 * John David Ratliff jdratlif@cs.indiana.edu
 * Chris O'Donnell chodonne@cs.indiana.edu
 *
 * lapd - the little asinine proxy daemon (a simple HTTP proxy)
 **************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "proxy.h"
#include "../mhttpd/queue.h"

// global data
short int running = TRUE;
vqueue_t *queue = NULL;
cache_t *cache = NULL;

// byte counts
unsigned long long bytes_from_cache = 0, bytes_from_http = 0;

// thread syncrhonization variables
pthread_mutex_t queue_mutex, server_mutex, cache_mutex;
pthread_cond_t full_cond, empty_cond, stop_cond;

// displays version information
void print_version_info(void) {
    printf("lapd - little asinine proxy daemon\n");
    printf("version %s\n", PROXY_VERSION);
}

// displays help and syntax screen
void print_help_info(void) {
    print_version_info();
    
    printf("\nsyntax: proxy < args >\n");
    printf("\t-h or --help\n\t\tDisplay this help screen.\n");
    printf("\t-v or --version\n\t\tDisplay version information.\n");
    printf("\t-p or --port\n\t\tThe port to run the server on.\n");
    printf("\t-t or --threads\n\t\tThe number of threads to use.\n");
  
    printf("\n");
}

// parses the command line arguments into SERVER_DATA format
void parse_command_line(const int argc, const char **argv, server_data_t *server) {
    int loop;
    const char *arg;
  
    // setup the default values
    server->port = DEFAULT_PORT;
    server->workers = DEFAULT_WORKERS;
    
    // loop through the argument list
    for (loop = 1; loop < argc; loop++) {
	arg = argv[loop];
    
	if ((strcmp(arg,"-p") == 0) || 
	    (strcmp(arg,"--port") == 0)) {
	    // server port argument
      
	    if ((++loop) < argc) {
		// save the port number
		server->port = atoi(argv[loop]);
	    }
	} else if ((strcmp(arg,"-h") == 0) || 
		   (strcmp(arg,"--help") == 0)) {
	    // help screen
	    print_help_info();
	    exit(EXIT_SUCCESS);
	} else if ((strcmp(arg,"-v") == 0) || 
		   (strcmp(arg,"--version") == 0)) {
	    // version info screen
	    print_version_info();
	    exit(EXIT_SUCCESS);
	} else if ((strcmp(arg,"-t") == 0) || 
		   (strcmp(arg,"--threads") == 0)) {
	    // worker threads to create
	    if ((++loop) < argc) {
		// the number of worker threads
		server->workers = atoi(argv[loop]);
	    }
	} else {
	    fprintf(stderr,"%s: invalid argument %s\n", argv[0], arg);
	}
    }
}

short int start_server(const uint16_t port, int *sockfd) {
    int server_sockfd = 0, yes = 1;
    struct sockaddr_in server_address;
  
    // create the server socket
    if ((server_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	perror("socket");
	return FAILURE;
    }
  
    // clear out the server_address structure
    memset(&server_address, 0, sizeof(struct sockaddr_in));
  
    // setup the server address structure
    server_address.sin_family = AF_INET;
    server_address.sin_addr.s_addr = INADDR_ANY;
    server_address.sin_port = htons(port);
  
    // allow the server's port to rebind
    if (setsockopt(server_sockfd, SOL_SOCKET, SO_REUSEADDR, 
		   &yes, sizeof(yes)) == -1) {
	perror("setsockopt");
    
	// make sure to close our socket
	close(server_sockfd);
    
	return FAILURE;
    }
  
    // attempt to bind the server's address to the port
    if (bind(server_sockfd, (struct sockaddr *)&server_address, 
	     sizeof(struct sockaddr)) == -1) {
	perror("bind");
    
	// make sure close the socket
	close(server_sockfd);
    
	return FAILURE;
    }
  
    // set the socket to listen for incoming connections
    if (listen(server_sockfd, 10) == -1) {
	perror("listen");
    
	// make sure to close our socket
	close(server_sockfd);
    
	return FAILURE;
    }
  
    // assign our server sockfd
    (*sockfd) = server_sockfd;
  
    return SUCCESS;
}

// wait on the server socket for incoming connections
void wait_for_connections(int sockfd) {
    struct sockaddr_in client_name;
    int client_size = sizeof(struct sockaddr_in);
    int client;
    int *client_ptr = NULL;
  
    // loop for...ever
    for (;;) {
	// accept a client connection
	if ((client = accept(sockfd, (struct sockaddr *)&client_name, 
			     &client_size)) == -1) {
	    if (!running) {
		break;
	    }
	}
    
	// add the client socket to the socket queue
	if ((client_ptr = (int *)malloc(sizeof(int))) == NULL) {
	    fprintf(stderr,"insufficient memory for socket pointer.\n");
	    close(client);
	    continue;
	}
    
	// save the client socket in the new memory block
	*client_ptr = client;
    
	// safeguard our queue
	pthread_mutex_lock(&queue_mutex);
    
	// wait on a full queue
	while ((running) && queue_full(queue)) {
	    pthread_cond_wait(&full_cond,&queue_mutex);
	}
    
	// if we're done running, end the loop
	if (!running) {
	    pthread_mutex_unlock(&queue_mutex);
	    break;
	}
    
	// enqueue the client socket pointer
	queue_enqueue(queue,client_ptr);
    
	// should be safe to unlock the mutex now
	pthread_mutex_unlock(&queue_mutex);
    
	// signal the queue is not empty
	pthread_cond_signal(&empty_cond);
    }
}

// the boss server thread -- just wait for connections
void *run_server(void *sockfd_ptr) {
    // wait for incoming connections
    wait_for_connections(*(int *)sockfd_ptr);
  
    // exit this thread
    pthread_exit(NULL);
  
    // will never reach this point
    return NULL;
}

// respond to a client HTTP/1.0 request
void respond_to_client(int sockfd) {
    char request[1024];
    size_t cache, http;
  
    // clear the request buffer
    memset(request, 0, 1024);
  
    // read the request from the client
    if (recv(sockfd, request, 1023, 0) == -1) {
	perror("recv");
    
	// make sure to close the client's socket
	close(sockfd);
    
	return;
    }
  
    // parse and respond, then close the socket
    parse(sockfd,request,&cache,&http);
    close(sockfd);

    // update the byte counts
    bytes_from_cache += cache;
    bytes_from_http += http;
}

// handle the connection requests
void *run_worker(void *unused) {
    int *sockfd_ptr = NULL;
    int sockfd;
  
    // loop for...ever
    for (;;) {
	// safeguard our queue
	pthread_mutex_lock(&queue_mutex);
    
	// wait on the empty queue
	while ((running) && queue_empty(queue)) {
	    pthread_cond_wait(&empty_cond,&queue_mutex);
	}
    
	if (!running) {
	    // if the server has stopped, break out of this thread
	    pthread_mutex_unlock(&queue_mutex);
	    break;
	}
    
	// dequeue our client connection socket
	queue_dequeue(queue,(void **)&sockfd_ptr);
    
	// should be safe to release the mutex now
	pthread_mutex_unlock(&queue_mutex);
    
	// signal the queue is not full
	pthread_cond_signal(&full_cond);
    
	// get a local copy and free the ptr memory
	sockfd = *sockfd_ptr;
	free(sockfd_ptr);
    
	// handle the connection...
	respond_to_client(sockfd);
    }
  
    // exit the thread
    pthread_exit(NULL);
  
    // will never reach here...
    return NULL;
}

// print the benchmark
void *run_benchmark(void *unused) {
    float f;

    for (;;) {
	// sleep 10 seconds...
	sleep(10);
	
	printf("bytes served from cache = %llu\n", bytes_from_cache);
	printf("bytes fetched from server = %llu\n", bytes_from_http);

	if ((bytes_from_cache + bytes_from_http) == 0) {
	    f = 0.0;
	} else {
	    f = ((float)bytes_from_cache / (float)(bytes_from_cache + bytes_from_http)) * 100;
	}

	printf("Amount served from cache: %%%f\n\n", f);

	if (!running) {
	    break;
	}
    }

    pthread_exit(NULL);

    // will never reach here...
    return NULL;
}

// returns the network byte ordered IP address of the host 'hostname'
// this one is thread-safe now!
in_addr_t get_ip_address(const char *hostname) {
    struct addrinfo *res = NULL;
    struct sockaddr_in *address;
    int error;
    in_addr_t ip = 0;

    error = getaddrinfo(hostname,NULL,NULL,&res);

    if (error == 0) {
	address = (struct sockaddr_in *)res->ai_addr;
	ip = address->sin_addr.s_addr;
	freeaddrinfo(res);
    }

    return ip;
}

// make a server connection
short int connect_to_server(const in_addr_t ip, const uint16_t port, int *sockfd) {
    int server_sockfd;
    struct sockaddr_in server;

    // clear the socket structure
    memset(&server, 0, sizeof(struct sockaddr_in));
	
    // setup the server's address
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = ip;
    server.sin_port = port;

    // open a socket fd
    if ((server_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	perror("socket");
	return FAILURE;
    }
	
    // connect to the server
    if (connect(server_sockfd, (const struct sockaddr *)&server, sizeof(struct sockaddr)) == -1) {
	perror("connect");
		
	// close the open socket
	close(server_sockfd);
		
	return FAILURE;
    }

    // assign our local data to our return data
    *sockfd = server_sockfd;
	
    return SUCCESS;
}

short int send_request(int http_sockfd, int client_sockfd, const char *request, char *cache, 
		       const int max_cache, size_t *bytes) {
    char response[8192];
    int length, read;
    int offset = 0;
	
    // clear the cache buffer
    memset(cache, 0, max_cache);
	
    // format the file request
    length = strlen(request);

    // reset the byte count
    *bytes = 0;
	
    // send the request to the server
    if (send(http_sockfd, request, length, 0) == -1) {
	perror("send");
		
	// close the http socket
	close(http_sockfd);
		
	return FAILURE;
    }
	
    // read the response
    do {
	// clear the response buffer
	memset(response, 0, 8192);

	// read from the server in 8K increments
	if ((read = recv(http_sockfd, response, 8192, 0)) == -1) {
	    perror("recv");
			
	    // close the http socket
	    close(http_sockfd);
			
	    return FAILURE;
	}
	
	if ((offset + read) < max_cache) {
	    // append the cache
	    memcpy(cache + offset, response, read);
	    offset += read;
	}

	// update the bytes read count
	*bytes += read;

	// send this part to the client
	if (send(client_sockfd, response, read, 0) == -1) {
	    perror("send");

	    // close the http socket
	    close(http_sockfd);

	    return FAILURE;
	}
    } while (read > 0);

    // close the http server socket
    close(http_sockfd);
	
    return SUCCESS;
}

// do a clean server shutdown
void stop_server(int signal) {
    // server is no longer running
    running = FALSE;
  
    // wakeup all sleeping threads...
    // so they know the server has stopped
    pthread_cond_broadcast(&empty_cond);
    pthread_cond_broadcast(&full_cond);
    pthread_cond_broadcast(&stop_cond);
}

// we start here...obviously...
int main(int argc, char **argv) {
    int sockfd = 0, loop;
    pthread_t boss, benchmark;
    pthread_t *workers = NULL;
    server_data_t server_data;
  
    // parse the command line arguments
    parse_command_line(argc, (const char **)argv, &server_data);
  
    // create the queue
    if ((queue = queue_create(DEFAULT_QUEUE_SIZE)) == NULL) {
	fprintf(stderr,"main: unable to create queue...\n");
	exit(EXIT_FAILURE);
    }

    if ((cache = cache_create(DEFAULT_CACHE_SIZE)) == NULL) {
	fprintf(stderr, "main: unable to create cache...\n");
	exit(EXIT_FAILURE);
    }
  
    // allocate memory for the workers array
    if ((workers = (pthread_t *)malloc(server_data.workers * 
				       sizeof(pthread_t))) == NULL) {
	fprintf(stderr,"insufficient memory to allocate thread array...\n");
	exit(EXIT_FAILURE);
    }
  
    // start the server
    if (start_server(server_data.port, &sockfd) == FAILURE) {
	fprintf(stderr,"unable to start server. exiting...\n");
	exit(EXIT_FAILURE);
    }
  
    // initialize all the synchronization variables
    pthread_mutex_init(&queue_mutex,NULL);
    pthread_mutex_init(&server_mutex,NULL);
    pthread_mutex_init(&cache_mutex,NULL);
    pthread_cond_init(&full_cond,NULL);
    pthread_cond_init(&empty_cond,NULL);
    pthread_cond_init(&stop_cond,NULL);
  
    // handle the interrupt signal and do a clean shutdown
    signal(SIGINT, stop_server);
  
    // ignore broken pipes
    signal(SIGPIPE, SIG_IGN);

    // create the boss thread
    if (pthread_create(&boss, NULL, run_server, &sockfd)) {
	fprintf(stderr,"unable to create boss thread. exiting...\n");
	exit(EXIT_FAILURE);
    }

    if (pthread_create(&benchmark, NULL, run_benchmark, NULL)) {
	fprintf(stderr,"unable to create benchmark thread. exiting...\n");
	exit(EXIT_FAILURE);
    }
    
    // create the worker threads
    for (loop = 0; loop < server_data.workers; loop++) {
	if (pthread_create((workers + loop), NULL, run_worker, NULL)) {
	    fprintf(stderr,"unable to create thread %d...\n", (loop + 1));
	}
    }
    
    // pthread_cond_wait requires a mutex, so keep one just for us	
    pthread_mutex_lock(&server_mutex);
    
    // wait until we get the stop condition
    while (running) {
	pthread_cond_wait(&stop_cond,&server_mutex);
    }
    
    // unlock our fake mutex
    pthread_mutex_unlock(&server_mutex);
    
    // close the server socket
    close(sockfd);
  
    // destroy the queue
    queue_destroy(queue);
    
    // destroy the cache
    cache_destroy(cache);
  
    // destroy all the synchronization variables
    pthread_mutex_destroy(&queue_mutex);
    pthread_mutex_destroy(&server_mutex);
    pthread_mutex_destroy(&cache_mutex);
    pthread_cond_destroy(&full_cond);
    pthread_cond_destroy(&empty_cond);
    pthread_cond_destroy(&stop_cond);
  
    // free the memory used by the worker thread array
    free(workers);
  
    // exit this thread
    pthread_exit(NULL);
  
    // will never get here
    return 0;
}

/******************************************************************************
 *  LAPD v 1.0.0
 *  Copyright (C) 2002 John David Ratliff [jdratlif@cs.indiana.edu]
 *                     Chris O'Donnell    [chodonne@cs.indiana.edu]
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation, 
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/
