/*************************************************************
 * CSCI P436 - Operating Systems
 *
 * John David Ratliff jdratlif@cs.indiana.edu
 * Chris O'Donnell chodonne@cs.indiana.edu
 *
 * sbc - Server Benchmark Client
 **************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <pthread.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "client.h"

// server address
struct sockaddr_in server;

// is the client running
short int running = TRUE;

// total bytes received from client connections
unsigned long int total_bytes = 0;

// safeguard our global data
pthread_mutex_t mutex;

// displays version information
void print_version_info(void) {
    printf("sbc - Server Benchmark Client\n");
    printf("version %s\n", SBC_VERSION);
}

// displays help and syntax screen
void print_help_info(void) {
    print_version_info();
	
    printf("\nsyntax: client < args >\n");
    printf("\t-h or --help\n\t\tDisplay this help screen.\n");
    printf("\t-v or --version\n\t\tDisplay version information.\n");
    printf("\t-p or --port\n\t\tThe port for the http server.\n");
    printf("\t-t or --threads\n\t\tThe number of threads to use.\n");
    printf("\t-r or --requests\n\t\tHow many requests each thread makes.\n");
    printf("\t-s or --server\n\t\tThe hostname or IP of the http server.\n");
    printf("\t-x or --proxy\n\t\tThe hostname or IP of the proxy server.\n");
    printf("\t-o or --proxy-port\n\t\tThe port of the proxy server.\n");

    printf("\n");
}

// parse the command line arguments
void parse_command_line(const int argc, const char **argv, client_data_t *client) {
    int loop;
    const char *arg;
	
    // setup the client defaults
    client->server_ip = get_ip_address(DEFAULT_HOST);
    client->server_port = htons((uint16_t)DEFAULT_HTTP_PORT);
    client->proxy_ip = ~0;
    client->proxy_port = htons((uint16_t)DEFAULT_PROXY_PORT);
    client->threads = DEFAULT_THREADS;
    client->requests = DEFAULT_REQUESTS;

    sprintf(client->http_server, "%s", DEFAULT_HOST);
    sprintf(client->http_port, "%d", DEFAULT_HTTP_PORT);
	
    for (loop = 1; loop < argc; loop++) {
	arg = argv[loop];
		
	if ((strcmp(arg,"-h") == 0) || (strcmp(arg,"--help") == 0)) {
	    // help
	    print_help_info();
	    exit(EXIT_SUCCESS);
	} else if ((strcmp(arg,"-v") == 0) || (strcmp(arg,"--version") == 0)) {
	    // version info
	    print_version_info();
	    exit(EXIT_SUCCESS);
	} else if ((strcmp(arg,"-s") == 0) || (strcmp(arg,"--server") == 0)) {
	    // server hostname
			
	    if ((++loop) < argc) {
		client->server_ip = get_ip_address(argv[loop]);
		strcpy(client->http_server, argv[loop]);
	    } else {
		fprintf(stderr,"Error: No server hostname specified!\n");
	    }
	} else if ((strcmp(arg,"-p") == 0) || (strcmp(arg,"--port") == 0)) {
	    // server port
			
	    if ((++loop) < argc) {
		client->server_port = htons((uint16_t)atoi(argv[loop]));
		strcpy(client->http_port, argv[loop]);
	    } else {
		fprintf(stderr,"Error: No server port specified!\n");
	    }
	} else if ((strcmp(arg,"-x") == 0) || (strcmp(arg,"--proxy") == 0)) {
	    if ((++loop) < argc) {
		client->proxy_ip = get_ip_address(argv[loop]);
	    } else {
		fprintf(stderr,"Error: No proxy hostname specified!\n");
	    }
	} else if ((strcmp(arg,"-o") == 0) || (strcmp(arg,"--proxy-port") == 0)) {
	    if ((++loop) < argc) {
		client->proxy_port = htons((uint16_t)atoi(argv[loop]));
	    } else {
		fprintf(stderr,"Error: No proxy port specified!\n");
	    }
	} else if ((strcmp(arg,"-t") == 0) || (strcmp(arg,"--threads") == 0)) {
	    // client threads
			
	    if ((++loop) < argc) {
		client->threads = atoi(argv[loop]);
	    } else {
		fprintf(stderr,"Error: No threads specified!\n");
	    }
	} else if ((strcmp(arg,"-r") == 0) || (strcmp(arg,"--requests") == 0)) {
	    // thread requests
			
	    if ((++loop) < argc) {
		client->requests = atoi(argv[loop]);
	    } else {
		fprintf(stderr,
			"Error: No request value specified!\n");
	    }
	} else {
	    // invalid argument
	    fprintf(stderr,"Error: invalid argument %s\n",arg);
	}
    }
}

// returns the network byte ordered IP address of the host 'hostname'
// this one is thread-safe now!
in_addr_t get_ip_address(const char *hostname) {
    struct addrinfo *res = NULL;
    struct sockaddr_in *address;
    int error;
    in_addr_t ip = 0;

    error = getaddrinfo(hostname,NULL,NULL,&res);

    if (error == 0) {
	address = (struct sockaddr_in *)res->ai_addr;
	ip = address->sin_addr.s_addr;
	freeaddrinfo(res);
    }

    return ip;
}

// make a server connection
short int connect_to_server(const struct sockaddr *server, int *sockfd) {
    int server_sockfd;

    // open a socket fd
    if ((server_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	perror("socket");
	return FAILURE;
    }
	
    // connect to the server
    if (connect(server_sockfd, server, sizeof(struct sockaddr)) == -1) {
	perror("connect");
		
	// close the open socket
	close(server_sockfd);
		
	return FAILURE;
    }

    // assign our local data to our return data
    *sockfd = server_sockfd;
	
    return SUCCESS;
}

short int send_request(int sockfd, unsigned long int *bytes, client_data_t *client) {
    char request[1024], response[8192];
    int fileno;
    int length, read;
	
    // clear the request buffer
    memset(request, 0, 1024);
	
    // reset the bytes counter
    *bytes = 0;

    // get a random file number, protect thread-unsafe rand() call.
    pthread_mutex_lock(&mutex);
    fileno = (rand() % TEST_FILE_COUNT) + 1;
    pthread_mutex_unlock(&mutex);
	
    // format the file request
    if (client->proxy_ip != ~(unsigned long int)0) {
	// we are using a proxy, so send a fully qualified URL
	length = sprintf(request, "GET http://%s:%s/file%d.html HTTP/1.0\r\n\r\n", 
			 client->http_server, client->http_port, fileno);
    } else {
	// no proxy, use relative
	length = sprintf(request, "GET /file%d.html HTTP/1.0\r\n\r\n", fileno);
    }
	
    // send the request to the server
    if (send(sockfd, request, length, 0) == -1) {
	perror("send");
		
	// close the client socket
	close(sockfd);
		
	return FAILURE;
    }
	
    // read the response
    do {
	// clear the response buffer
	memset(response, 0, 8192);

	// read from the server in 8K increments
	if ((read = recv(sockfd, response, 8191, 0)) == -1) {
	    perror("recv");
			
	    // close the client socket
	    close(sockfd);
			
	    return FAILURE;
	}
		
	// update the byte count
	*bytes += read;
    } while (read > 0);
	
    return SUCCESS;
}

// make multiple requests to an HTTP/1.0 server
void *make_requests(void *client_ptr) {
    client_data_t *client = (client_data_t *)client_ptr;
    short int requests = client->requests;
    int loop, sockfd = 0;
    unsigned long int bytes = 0;

    // do 'requests' requests
    for (loop = 0; loop < requests; loop++) {
	// make sure we are still running
	if (!running) {
	    // if the client is no longer running, break out of the loop
	    break;
	}

	// make a server connection
	if (connect_to_server((const struct sockaddr *)&server, 
			      &sockfd) == FAILURE) {
	    fprintf(stderr,"worker: unable to connect to server...\n");
	    continue;
	}
		
	// send the HTTP/1.0 request
	if (send_request(sockfd, &bytes, client) == FAILURE) {
	    fprintf(stderr,"worker: unable to complete request...\n");
	    continue;
	}
		
	// close the client socket
	close(sockfd);
		
	// update the global byte counter
	pthread_mutex_lock(&mutex);
	total_bytes += bytes;
	pthread_mutex_unlock(&mutex);
    }
	
    // exit this thread
    pthread_exit(NULL);
	
    // will never reach here
    return NULL;
}

// informs all who care the client is not supposed to run anymore
// does not actually STOP anything. only changes the state
void stop_client(int signal) {
    // client is no longer running
    running = FALSE;
}

// start running here...
int main(int argc, char **argv) {
    client_data_t client_data;
    pthread_t *threads = NULL;
    int loop;
    void *ptr = NULL;
    time_t elapsed;
    unsigned long int kbps = 0;
    
    // parse the command line arguments
    parse_command_line(argc, (const char **)argv, &client_data);

    // seed the random number generator
    srand(time(NULL));
	
    // clear the socket structure
    memset(&server, 0, sizeof(struct sockaddr_in));
	
    // setup the server's address
    server.sin_family = AF_INET;

    if (client_data.proxy_ip != ~(unsigned long int)0) {
	server.sin_addr.s_addr = client_data.proxy_ip;
	server.sin_port = client_data.proxy_port;
    } else {
	server.sin_addr.s_addr = client_data.server_ip;
	server.sin_port = client_data.server_port;
    }
	
    // initialize the mutex variables
    pthread_mutex_init(&mutex,NULL);
	
    // handle the interrupt signal
    signal(SIGINT, stop_client);
	
    // allocate memory for the threads
    if ((threads = (pthread_t *)malloc(client_data.threads * 
				       sizeof(pthread_t))) == NULL) {
	fprintf(stderr,"Insufficient memory to create threads array...\n");
	exit(EXIT_FAILURE);
    }
	
    // grab the time before the threads start
    elapsed = time(NULL);
	
    // create the threads
    for (loop = 0; loop < client_data.threads; loop++) {
	if (pthread_create((threads + loop), NULL, 
			   make_requests, &client_data)) {
	    fprintf(stderr,"Unable to create thread %d...\n", (loop + 1));
	}
    }
	
    // wait for threads to exit
    for (loop = 0; loop < client_data.threads; loop++) {
	pthread_join(threads[loop], &ptr);
    }
	
    // calculate the elapsed time
    elapsed = time(NULL) - elapsed;
	
    printf("Total bytes received: %lu\n", total_bytes);
    printf("Time: ~%lu seconds.\n", elapsed);
	
    // calculate the average speed
    kbps = (total_bytes / 1024);
    kbps /= (elapsed > 0) ? elapsed : 1;
	
    printf("Average speed: %lu KB/s, %lu MB/s\n\n", kbps, (kbps / 1024));
		
    // destroy the mutex
    pthread_mutex_destroy(&mutex);
	
    // free the threads array
    free(threads);
	
    // exit the main thread
    pthread_exit(NULL);
	
    // will never reach here
    return 0;
}

/******************************************************************************
 *  Server Benchmark Client 1.2.0
 *  Copyright (C) 2002 John David Ratliff [jdratlif@cs.indiana.edu]
 *                     Chris O'Donnell    [chodonne@cs.indiana.edu]
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation, 
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/
